
//		Ruler.cpp - On-screen ruler for Take Command


#define UNICODE								1
#define _UNICODE							1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define PLUGIN_NAME							L"Ruler"

#define VER_MAJOR							1
#define VER_MINOR							0
#define VER_BUILD							2
#define VER_PATCH							4

#ifdef _WIN64
#define PLATFORM							L"x64"
#else
#define PLATFORM							L"x86"
#endif

#define MAX_ARG_LEN							8192
#define SHORT_BUF_LEN						1024

#define ERRHANDLE							GetStdHandle( STD_ERROR_HANDLE )

#define MAGIC_VALUE							0xfedcba98


//		Remove this to make the plugin smaller:
//		#define DEBUG_CODE


#define RULER_DEFAULT_HOTKEY				0x10088
#define RULER_DEFAULT_HOTKEY_MODE			1
#define RULER_DEFAULT_TRANSPMODE			0
#define RULER_DEFAULT_BACKCOLOR				0xe6e0b0
#define RULER_DEFAULT_HASHCOLOR				0x000000
#define RULER_DEFAULT_ALPHA					128
#define RULER_DEFAULT_PEN_WIDTH				1

#define RULER_NUM_HOTKEYS					5
#define RULER_HOTKEY_MODE_MAX				17

#define RULER_POSITION_TOP					0
#define RULER_POSITION_BOTTOM				1
#define RULER_POSITION_LEFT					2
#define RULER_POSITION_RIGHT				3

#define RULER_ALPHA_MIN						64
#define RULER_ALPHA_MAX						255
#define RULER_PEN_WIDTH_MAX					3

#define RULER_MAX_ALT_MARKS					16
#define RULER_MAX_COLUMN					2000
#define RULER_MAX_COLUMN_REPEAT				100
#define RULER_FLAG_MARK_REPEATS				0x10000

#define RULER_DEFAULT_TIMER_ID				1977
#define RULER_TIMER_TIMEOUT					100

#define RULER_MODE_CONSOLE					0
#define RULER_MODE_TCMD						1

#define DEFAULT_TCMD_HMARGIN				4
#define DEFAULT_TCMD_VMARGIN				2

#define MAX_COLOR							0x00ffffff



// --------------------  Global variables  --------------------


typedef BOOL ( WINAPI *GET_CSB_INFO_EX_PROC ) ( HANDLE hConsole, CONSOLE_SCREEN_BUFFER_INFOEX *BufferInfoEx );

HMODULE Kernel32Handle						= NULL;
HINSTANCE MyDLLHandle						= NULL;
HINSTANCE TakeCmdDllHandle					= NULL;
GET_CSB_INFO_EX_PROC GetCSBInfoExProc		= NULL;

wchar_t RulerWindowClass[SHORT_BUF_LEN]		= L"";

ATOM RulerWindowAtom						= 0;
HWND ConsoleRuler							= NULL;
HWND TakeCmdRuler							= NULL;

UINT_PTR TimerID							= 0;

HWND ConsoleWindow							= NULL;
HWND TakeCmdWindow							= NULL;
HWND MyTab									= NULL;
HWND DesktopWindow							= NULL;
BOOL LastVisible							= FALSE;
HMONITOR LastConsoleMonitor					= NULL;
RECT ClientArea								= { 0, 0, 0, 0 };

unsigned int DelayedRedraw					= 0;

#ifdef DEBUG_CODE
DWORD Debug									= FALSE;
#endif	//	DEBUG_CODE

HANDLE ConsoleHandle						= NULL;

int Hotkey[RULER_NUM_HOTKEYS]				= { RULER_DEFAULT_HOTKEY };
DWORD HotkeyMode[RULER_NUM_HOTKEYS]			= { RULER_DEFAULT_HOTKEY_MODE };
COLORREF BackColor							= RULER_DEFAULT_BACKCOLOR;
COLORREF HashColor							= RULER_DEFAULT_HASHCOLOR;
COLORREF RulerColor							= MAX_COLOR;
COLORREF AltColumnsColor					= MAX_COLOR;
COLORREF AltRowsColor						= MAX_COLOR;
DWORD Alpha									= RULER_DEFAULT_ALPHA;
DWORD PenWidth								= RULER_DEFAULT_PEN_WIDTH;
DWORD PenWidth2								= RULER_DEFAULT_PEN_WIDTH;
DWORD Position								= RULER_POSITION_TOP;
unsigned int TopOffset						= 0;
DWORD Disabled								= FALSE;
DWORD DisabledAtStartup						= FALSE;
DWORD TranspMode							= RULER_DEFAULT_TRANSPMODE;
DWORD HookListCmd							= TRUE;

CONSOLE_SCREEN_BUFFER_INFO	BufferInfo;
unsigned int TCmdHMargin					= DEFAULT_TCMD_HMARGIN;
unsigned int TCmdVMargin					= DEFAULT_TCMD_VMARGIN;
unsigned int ViewportWidth					= 0;
unsigned int ViewportHeight					= 0;

unsigned int TCCVerMajor					= 0;
unsigned int TCCVerMinor					= 0;

unsigned int NumAltColumns					= 0;
unsigned int AltColumns[RULER_MAX_ALT_MARKS];

unsigned int NumAltRows						= 0;
unsigned int AltRows[RULER_MAX_ALT_MARKS];



COLORREF RulerColorForBG[16]				= { 0xffbf00, 0xffbf00, 0x98fb98, 0xdcdcdc,
												0x507fff, 0xffffff, 0x000000, 0x4f4f2f, 
												0x000000, 0x000000, 0x000000, 0x000000, 
												0x000000, 0x000000, 0x008cff, 0xa09e5f };


const COLORREF DefaultColorForBG[16]		= { 0xffbf00, 0xffbf00, 0x98fb98, 0xdcdcdc,
												0x507fff, 0xffffff, 0x000000, 0x4f4f2f, 
												0x000000, 0x000000, 0x000000, 0x000000, 
												0x000000, 0x000000, 0x008cff, 0xa09e5f };


COLORREF CurrentConsoleColors[16]			= { 0x000000, 0x800000, 0x008000, 0x808000,
												0x000080, 0x800080, 0x008080, 0xc0c0c0,
												0x808080, 0xff0000, 0x00ff00, 0xffff00,
												0x0000ff, 0xff00ff, 0x00ffff, 0xffffff };


const COLORREF DefaultConsoleColors[16]		= { 0x000000, 0x800000, 0x008000, 0x808000,
												0x000080, 0x800080, 0x008080, 0xc0c0c0,
												0x808080, 0xff0000, 0x00ff00, 0xffff00,
												0x0000ff, 0xff00ff, 0x00ffff, 0xffffff };


// ------------------------------------------------------------



struct _RulerWinData {
	unsigned int Mode;
	HWND Parent;
	unsigned int HMargin;
	unsigned int VMargin;
	unsigned int HiddenRedrawCount;
	BOOL DesktopHack;
};



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetLastErrorText( LPTSTR outString )
{
	return GetErrorText( GetLastError(), outString );
}



#include "ParseArgs.cpp"
#include "HotKeys.cpp"
#include "w3ccolors.cpp"



wchar_t PluginFeatures[] = L"ruler,list,*key";



const wchar_t RulerHelpText[] = L"View and set ruler options.\n\
\n\
RULER /A: /B: /C: /D /E: /H: /K: /L /M: /P: /Q /R /S /U: /W: /X: /Z: \r\n\
\n\
\t/A:\tset the ruler\u2019s alpha transparency (alpha mode)  \n\
\t/B:\tset the ruler\u2019s background color (alpha mode)  \n\
\t/C:\tset the ruler color (color-key mode)\n\
\t/D\trestore default plugin settings\n\
\t/E:\tenable or disable (show or hide) the ruler\n\
\t/H:\tset the color of the hash marks (alpha mode)\n\
\t/K:\tset the hotkey\n\
\t/L\tlist all supported hotkeys\n\
\t/M:\tset the hotkey mode\n\
\t/P:\tset the ruler\u2019s position\n\
\t/Q\tquietly\n\
\t/R\trestore saved settings from the registry\n\
\t/S\tsave the current settings to the registry\n\
\t/U:\tintercept the LIST command\n\
\t/W:\tset the width of the hash marks\n\
\t/X:\tset the ruler\u2019s transparency mode\n\
\t/Z:\tset whether the ruler is hidden at startup\n\
\n\
RULER with no options displays the plugin\u2019s current settings.  \n";



const int NumHelpTexts							= 1;
const wchar_t HelpFeature[NumHelpTexts][8]		= { L"ruler" };
const wchar_t *HelpText[NumHelpTexts]			= { RulerHelpText };



const wchar_t PluginRegKey[]					= L"Software\\JPPlugins\\Ruler";
const wchar_t HotKeyRegValue[]					= L"Hotkey";
const wchar_t HotKeyNRegValue[]					= L"Hotkey #%u";
const wchar_t HotkeyModeRegValue[]				= L"Hotkey Mode";
const wchar_t HotkeyNModeRegValue[]				= L"Hotkey Mode #%u";
const wchar_t BackColorRegValue[]				= L"Background Color";
const wchar_t HashColorRegValue[]				= L"Hash Color";
const wchar_t AlphaRegValue[]					= L"Alpha";
const wchar_t PenWidthRegValue[]				= L"Pen Width";
const wchar_t PenWidth2RegValue[]				= L"Pen Width 2";
const wchar_t PositionRegValue[]				= L"Position";
const wchar_t DisabledAtStartRegValue[]			= L"Hidden at Startup";
const wchar_t TranspModeRegValue[]				= L"Transparency Mode";
const wchar_t RulerColorRegValue[]				= L"Ruler Color %02u";
const wchar_t NumAltColumnsRegValue[]			= L"Alt Columns Count";
const wchar_t AltColumnsRegValue[]				= L"Alt Column #%u";
const wchar_t AltColumnsColorRegValue[]			= L"Alt Columns Color";
const wchar_t NumAltRowsRegValue[]				= L"Alt Rows Count";
const wchar_t AltRowsRegValue[]					= L"Alt Row #%u";
const wchar_t AltRowsColorRegValue[]			= L"Alt Rows Color";
const wchar_t HookListCmdRegValue[]				= L"Hook LIST";

#ifdef DEBUG_CODE
const wchar_t DebugRegValue[]					= L"Debug";
#endif	//	DEBUG_CODE


const wchar_t TranspModeStr[2][10]				= { L"Color key", L"Alpha" };
const wchar_t PositionStr[4][7]					= { L"Top", L"Bottom", L"Left", L"Right" };

const wchar_t HotkeyModeStr[RULER_HOTKEY_MODE_MAX + 1][30]	
												= { L"Enable/Disable", L"Top/Bottom/Disable", L"Top/Bottom/Left/Disable",
													L"Top/Bottom/Left/Right/Disable", L"Top/Bottom", L"Top/Bottom/Left",
													L"Top/Bottom/Left/Right", L"Top/Left/Disable", L"Top/Left",
													L"Enable", L"Disable", L"Top", L"Bottom", L"Left", L"Right",
													L"Left/Right/Disable", L"Left/Right", L"ColorKey/Alpha" };



BOOL IsColonOrEquals( wchar_t ch )
{
	if ( ( ch == ':' ) || ( ch == '=' ) )
		return TRUE;
	else
		return FALSE;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value	= 0;
	wchar_t *sep1		= NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



#ifdef DEBUG_CODE

LPCTSTR RulerModeStr( unsigned int Mode )
{
	static const wchar_t ConsoleMode[]		= L"Console";
	static const wchar_t TakeCmdMode[]		= L"TakeCmd";
	static const wchar_t OtherMode[]		= L"Unknown";

	
	if ( Mode == RULER_MODE_CONSOLE )
		return ConsoleMode;
	else if ( Mode == RULER_MODE_TCMD )
		return TakeCmdMode;
	else
		return OtherMode;
}



LPCTSTR RulerModeStr( _RulerWinData *Data )
{
	static wchar_t shortbuf[SHORT_BUF_LEN]		= L"";


	if ( Data )
		wsprintf( shortbuf, L"(%s ruler)", RulerModeStr( Data->Mode ) );
	else
		shortbuf[0] = '\0';

	return shortbuf;
}

#endif	//	DEBUG_CODE



int GetTCCVer()
{
	wchar_t temp[SHORT_BUF_LEN]		= L"%_4ver";

	unsigned int Major				= 0;
	unsigned int Minor				= 0;
	wchar_t *ch						= NULL;


	ExpandVariables( temp, 1 );


	_set_errno( 0 );
	Major = wcstoul( temp, &ch, 10 );

	if ( errno )
		Major = 0;

	while ( ( *ch == '.' ) || ( *ch == ',' ) )
		ch++;

	if ( iswdigit( *ch ) ) {
		_set_errno( 0 );
		Minor = wcstoul( ch, NULL, 10 );

		if ( errno )
			Minor = 0;
	}

	TCCVerMajor = Major;
	TCCVerMinor = Minor;
	return 0;
}



BOOL IsAFile( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;


	DWORD attr		= GetFileAttributes( Filename );


	if ( attr == INVALID_FILE_ATTRIBUTES )
		return FALSE;

	if ( attr & ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE ) )
		return FALSE;

	return TRUE;
}



//		Attempts to read the HMargin and VMargin values from Take Command's .INI file.
//		(These values first appeared in version 24.)  On any error, default values
//		will be returned.

int GetTCmdMargins()
{
	wchar_t IniFile[MAX_ARG_LEN]			= L"%_ininame";
	wchar_t HMarginValue[SHORT_BUF_LEN]		= L"";
	wchar_t VMarginValue[SHORT_BUF_LEN]		= L"";


	TCmdHMargin		= DEFAULT_TCMD_HMARGIN;
	TCmdVMargin		= DEFAULT_TCMD_VMARGIN;

	ExpandVariables( IniFile, 1 );

	if ( IniFile[0] == '\0' )
		return 1;

	if ( !IsAFile( IniFile ) )		//	if ( !QueryIsFile( IniFile ) )
		return 2;

	GetPrivateProfileString( L"TakeCommand", L"HMargin", NULL, HMarginValue, SHORT_BUF_LEN, IniFile );
	GetPrivateProfileString( L"TakeCommand", L"VMargin", NULL, VMarginValue, SHORT_BUF_LEN, IniFile );

	if ( iswdigit( HMarginValue[0] ) ) {
		_set_errno( 0 );
		TCmdHMargin = wcstoul( HMarginValue, NULL, 10 );

		if ( errno )
			TCmdHMargin = DEFAULT_TCMD_HMARGIN;
	}

	if ( iswdigit( VMarginValue[0] ) ) {
		_set_errno( 0 );
		TCmdVMargin = wcstoul( VMarginValue, NULL, 10 );

		if ( errno )
			TCmdVMargin = DEFAULT_TCMD_VMARGIN;
	}

	return 0;
}



//		Restore the built-in default values :

int RestorePluginDefaults()
{
	memset( Hotkey, 0x00, RULER_NUM_HOTKEYS * sizeof( int ) );
	memset( HotkeyMode, 0x00, RULER_NUM_HOTKEYS * sizeof( DWORD ) );

	Hotkey[0]			= RULER_DEFAULT_HOTKEY;
	HotkeyMode[0]		= RULER_DEFAULT_HOTKEY_MODE;
	BackColor			= RULER_DEFAULT_BACKCOLOR;
	HashColor			= RULER_DEFAULT_HASHCOLOR;
	Alpha				= RULER_DEFAULT_ALPHA;
	PenWidth			= RULER_DEFAULT_PEN_WIDTH;
	PenWidth2			= RULER_DEFAULT_PEN_WIDTH;
	Position			= RULER_POSITION_TOP;
	TopOffset			= 0;
	DisabledAtStartup	= FALSE;
	TranspMode			= RULER_DEFAULT_TRANSPMODE;
	NumAltColumns		= 0;
	AltColumnsColor		= MAX_COLOR;
	NumAltRows			= 0;
	AltRowsColor		= MAX_COLOR;
	HookListCmd			= TRUE;

	for ( int i = 0; i < 16; i++ )
		RulerColorForBG[i] = DefaultColorForBG[i];

	return 0;
}



BOOL GoodAltMark( unsigned int Mark )
{
	if ( ( Mark >= 1 ) && ( Mark <= RULER_MAX_COLUMN ) )
		return TRUE;

	if ( ( Mark >= ( RULER_FLAG_MARK_REPEATS | 2 ) ) && ( Mark <= ( RULER_FLAG_MARK_REPEATS | RULER_MAX_COLUMN_REPEAT ) ) )
		return TRUE;

	return FALSE;
}



DWORD RegGetDWord( HKEY KeyHandle, LPCTSTR ValueName, DWORD Min, DWORD Max, DWORD Default, int *Error )
{
	DWORD value			= -1;
	DWORD bufsize		= sizeof( DWORD );
	DWORD valtype		= 0;


	RegQueryValueEx( KeyHandle, ValueName, NULL, &valtype, (BYTE*) &value, &bufsize );

	if ( ( valtype != REG_DWORD ) || ( value < Min ) || ( value > Max ) ) {
		value = Default;

		if ( Error )
			*Error += 1;
	}

	return value;
}



//		Restore saved options from the registry :

int RestoreOptions( HKEY RootKey )
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	HKEY KeyHandle			= NULL;
	DWORD hotkeyval			= -1;
	DWORD keymodeval		= -1;
	DWORD transpmodeval		= -1;
	DWORD backcolorval		= -1;
	DWORD hashcolorval		= -1;
	DWORD alphaval			= -1;
	DWORD penwidthval		= -1;
	DWORD penwidth2val		= -1;
	DWORD positionval		= -1;
	DWORD startdisabledval	= -1;
	DWORD numaltcolumnsval	= -1;
	DWORD numaltrowsval		= -1;
	DWORD altcolscolorval	= -1;
	DWORD altrowscolorval	= -1;
	DWORD hooklistcmdval	= -1;

#ifdef DEBUG_CODE
	DWORD debugval			= -1;
#endif	//	DEBUG_CODE

	DWORD altcolumnsval[RULER_MAX_ALT_MARKS];
	DWORD altrowsval[RULER_MAX_ALT_MARKS];

	int rv					= 0;


	if ( RegOpenKeyEx( RootKey, PluginRegKey, 0, KEY_READ, &KeyHandle ) != ERROR_SUCCESS )
		return 1;


	hotkeyval			= RegGetDWord( KeyHandle, HotKeyRegValue, 0, UINT_MAX, UINT_MAX, NULL );
	if ( hotkeyval && !IsKnownKeyValue( hotkeyval ) )
		rv++;

	keymodeval			= RegGetDWord( KeyHandle, HotkeyModeRegValue, 0, RULER_HOTKEY_MODE_MAX, RULER_DEFAULT_HOTKEY_MODE, NULL );
	transpmodeval		= RegGetDWord( KeyHandle, TranspModeRegValue, 0, 1, RULER_DEFAULT_TRANSPMODE, NULL );
	backcolorval		= RegGetDWord( KeyHandle, BackColorRegValue, 0, MAX_COLOR, RULER_DEFAULT_BACKCOLOR, NULL );
	hashcolorval		= RegGetDWord( KeyHandle, HashColorRegValue, 0, MAX_COLOR, RULER_DEFAULT_HASHCOLOR, NULL );
	alphaval			= RegGetDWord( KeyHandle, AlphaRegValue, RULER_ALPHA_MIN, RULER_ALPHA_MAX, RULER_DEFAULT_ALPHA, NULL );
	penwidthval			= RegGetDWord( KeyHandle, PenWidthRegValue, 0, RULER_PEN_WIDTH_MAX, RULER_DEFAULT_PEN_WIDTH, NULL );
	penwidth2val		= RegGetDWord( KeyHandle, PenWidth2RegValue, 0, RULER_PEN_WIDTH_MAX, penwidthval, NULL );
	positionval			= RegGetDWord( KeyHandle, PositionRegValue, RULER_POSITION_TOP, RULER_POSITION_RIGHT, RULER_POSITION_TOP, NULL );
	startdisabledval	= RegGetDWord( KeyHandle, DisabledAtStartRegValue, FALSE, TRUE, FALSE, NULL );
	numaltcolumnsval	= RegGetDWord( KeyHandle, NumAltColumnsRegValue, 0, RULER_MAX_ALT_MARKS, 0, NULL );
	numaltrowsval		= RegGetDWord( KeyHandle, NumAltRowsRegValue, 0, RULER_MAX_ALT_MARKS, 0, NULL );
	altcolscolorval		= RegGetDWord( KeyHandle, AltColumnsColorRegValue, 0, MAX_COLOR, MAX_COLOR, NULL );
	altrowscolorval		= RegGetDWord( KeyHandle, AltRowsColorRegValue, 0, MAX_COLOR, MAX_COLOR, NULL );
	hooklistcmdval		= RegGetDWord( KeyHandle, HookListCmdRegValue, 0, UINT_MAX, TRUE, NULL );

#ifdef DEBUG_CODE

	debugval			= RegGetDWord( KeyHandle, DebugRegValue, FALSE, TRUE, FALSE, NULL );

#endif	//	DEBUG_CODE


	for ( unsigned int i = 0; i < numaltcolumnsval; i++ ) {
		wsprintf( temp, AltColumnsRegValue, i + 1 );

		altcolumnsval[i] = RegGetDWord( KeyHandle, temp, 0, UINT_MAX, UINT_MAX, &rv );
		if ( !GoodAltMark( altcolumnsval[i] ) )
			rv++;
	}


	for ( unsigned int i = 0; i < numaltrowsval; i++ ) {
		wsprintf( temp, AltRowsRegValue, i + 1 );

		altrowsval[i]	= RegGetDWord( KeyHandle, temp, 0, UINT_MAX, UINT_MAX, &rv );
		if ( !GoodAltMark( altrowsval[i] ) )
			rv++;
	}


	if ( !rv ) {
		Hotkey[0]			= hotkeyval;
		HotkeyMode[0]		= keymodeval;
		TranspMode			= transpmodeval;
		BackColor			= backcolorval;
		HashColor			= hashcolorval;
		Alpha				= alphaval;
		PenWidth			= penwidthval;
		PenWidth2			= penwidth2val;
		Position			= positionval;
		TopOffset			= 0;
		DisabledAtStartup	= startdisabledval;
		NumAltColumns		= 0;
		NumAltRows			= 0;
		AltColumnsColor		= altcolscolorval;
		AltRowsColor		= altrowscolorval;
		HookListCmd			= hooklistcmdval;

#ifdef DEBUG_CODE

		Debug				= debugval;

#endif	//	DEBUG_CODE

		for ( int i = 0; i < 16; i++ ) {
			wsprintf( temp, RulerColorRegValue, i );
			RulerColorForBG[i] = RegGetDWord( KeyHandle, temp, 0, MAX_COLOR, DefaultColorForBG[i], NULL );
		}

		if ( numaltcolumnsval ) {
			for ( unsigned int i = 0; i < numaltcolumnsval; i++ )
				AltColumns[i] = altcolumnsval[i];

			NumAltColumns = numaltcolumnsval;
		}

		if ( numaltrowsval ) {
			for ( unsigned int i = 0; i < numaltrowsval; i++ )
				AltRows[i] = altrowsval[i];

			NumAltRows = numaltrowsval;
		}

		for ( int i = 1; i < RULER_NUM_HOTKEYS; i++ ) {
			wsprintf( temp, HotKeyNRegValue, i + 1 );
			Hotkey[i] = RegGetDWord( KeyHandle, temp, 0, UINT_MAX, UINT_MAX, NULL );

			wsprintf( temp, HotkeyNModeRegValue, i + 1 );
			HotkeyMode[i] = RegGetDWord( KeyHandle, temp, 0, UINT_MAX, UINT_MAX, NULL );

			if ( !IsKnownKeyValue( Hotkey[i] ) || HotkeyMode[i] > RULER_HOTKEY_MODE_MAX )
				Hotkey[i] = HotkeyMode[i] = 0;
		}
	}

	RegCloseKey( KeyHandle );
	return rv;
}



int RestoreOptions()
{
	if ( RestoreOptions( HKEY_CURRENT_USER ) == 0 )
		return 1;

	if ( RestoreOptions( HKEY_LOCAL_MACHINE ) == 0 )
		return 2;

	return 0;
}



int CondenseHotkeys()
{
	int f			= 0;
	int n			= 0;


	//		Remove duplicate key entries :

	for ( int i = 0; i < ( RULER_NUM_HOTKEYS - 1 ); i++ ) {
		if ( Hotkey[i] ) {
			f = 0;
			for ( int j = i + 1; j < RULER_NUM_HOTKEYS; j++ ) {
				if ( Hotkey[j] == Hotkey[i] ) {
					Hotkey[j]		= 0;
					HotkeyMode[j]	= 0;
					n++;
				}
			}
		}
	}


	//		Remove 'air' from list of hotkeys :

	for ( int i = 0; i < ( RULER_NUM_HOTKEYS - 1 ); i++ ) {
		if ( Hotkey[i] == 0 ) {
			for ( int j = i + 1; j < RULER_NUM_HOTKEYS; j++ ) {
				if ( Hotkey[j] ) {
					Hotkey[i]		= Hotkey[j];
					HotkeyMode[i]	= HotkeyMode[j];
					Hotkey[j]		= 0;
					HotkeyMode[j]	= 0;
					n++;
					break;
				}
			}
		}
	}

	return n;
}



int SaveOptions( HKEY RootKey )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	HKEY KeyHandle			= NULL;

	int t					= 0;
	int rv					= 0;


	if ( ( t = RegCreateKeyEx( RootKey, PluginRegKey, 0, NULL, 0, KEY_READ | KEY_WRITE, NULL, &KeyHandle, NULL ) ) == ERROR_SUCCESS ) {

		if ( t = RegSetValueEx( KeyHandle, HotKeyRegValue, 0, REG_DWORD, (BYTE*) Hotkey, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, HotkeyModeRegValue, 0, REG_DWORD, (BYTE*) &HotkeyMode, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, TranspModeRegValue, 0, REG_DWORD, (BYTE*) &TranspMode, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, BackColorRegValue, 0, REG_DWORD, (BYTE*) &BackColor, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, HashColorRegValue, 0, REG_DWORD, (BYTE*) &HashColor, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, AlphaRegValue, 0, REG_DWORD, (BYTE*) &Alpha, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, PenWidthRegValue, 0, REG_DWORD, (BYTE*) &PenWidth, sizeof( DWORD ) ) )
			rv = t;

		if ( PenWidth != PenWidth2 ) {
			if ( t = RegSetValueEx( KeyHandle, PenWidth2RegValue, 0, REG_DWORD, (BYTE*) &PenWidth2, sizeof( DWORD ) ) )
				rv = t;
		}
		else
			RegDeleteValue( KeyHandle, PenWidth2RegValue );

		if ( t = RegSetValueEx( KeyHandle, PositionRegValue, 0, REG_DWORD, (BYTE*) &Position, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, DisabledAtStartRegValue, 0, REG_DWORD, (BYTE*) &DisabledAtStartup, sizeof( DWORD ) ) )
			rv = t;

		for ( int i = 0; i < 16; i++ ) {
			wsprintf( temp, RulerColorRegValue, i );

			if ( RulerColorForBG[i] == DefaultColorForBG[i] )
				RegDeleteValue( KeyHandle, temp );
			else if ( t = RegSetValueEx( KeyHandle, temp, 0, REG_DWORD, (BYTE*) ( RulerColorForBG + i ), sizeof( DWORD ) ) )
				rv = t;
		}

		if ( t = RegSetValueEx( KeyHandle, NumAltColumnsRegValue, 0, REG_DWORD, (BYTE*) &NumAltColumns, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, NumAltRowsRegValue, 0, REG_DWORD, (BYTE*) &NumAltRows, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, AltColumnsColorRegValue, 0, REG_DWORD, (BYTE*) &AltColumnsColor, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, AltRowsColorRegValue, 0, REG_DWORD, (BYTE*) &AltRowsColor, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, HookListCmdRegValue, 0, REG_DWORD, (BYTE*) &HookListCmd, sizeof( DWORD ) ) )
			rv = t;

		for ( unsigned int i = 0; i < RULER_MAX_ALT_MARKS; i++ ) {
			wsprintf( temp, AltColumnsRegValue, i + 1 );

			if ( i < NumAltColumns ) {
				if ( t = RegSetValueEx( KeyHandle, temp, 0, REG_DWORD, (BYTE*) ( AltColumns + i ), sizeof( DWORD ) ) )
					rv = t;
			}
			else
				RegDeleteValue( KeyHandle, temp );

			wsprintf( temp, AltRowsRegValue, i + 1 );

			if ( i < NumAltRows ) {
				if ( t = RegSetValueEx( KeyHandle, temp, 0, REG_DWORD, (BYTE*) ( AltRows + i ), sizeof( DWORD ) ) )
					rv = t;
			}
			else
				RegDeleteValue( KeyHandle, temp );
		}

		for ( int i = 1; i < RULER_NUM_HOTKEYS; i++ ) {
			wsprintf( temp, HotKeyNRegValue, i + 1 );
			wsprintf( temp + 256, HotkeyNModeRegValue, i + 1 );

			if ( Hotkey[i] ) {
				if ( t = RegSetValueEx( KeyHandle, temp, 0, REG_DWORD, (BYTE*) ( Hotkey + i ), sizeof( DWORD ) ) )
					rv = t;

				if ( t = RegSetValueEx( KeyHandle, temp + 256, 0, REG_DWORD, (BYTE*) ( HotkeyMode + i ), sizeof( DWORD ) ) )
					rv = t;
			}
			else {
				RegDeleteValue( KeyHandle, temp );
				RegDeleteValue( KeyHandle, temp + 256 );
			}
		}

		RegCloseKey( KeyHandle );
	}
	else
		rv = t;

	return rv;
}



int SaveStartupOptions( HKEY RootKey, int HiddenAtStartup, int HookList )
{
	HKEY KeyHandle			= NULL;

	DWORD hotkeyval			= 0;

	int t					= 0;
	int rv					= 0;



	if ( ( t = RegCreateKeyEx( RootKey, PluginRegKey, 0, NULL, 0, KEY_READ | KEY_WRITE, NULL, &KeyHandle, NULL ) ) == ERROR_SUCCESS ) {

		if ( HiddenAtStartup >= 0 )
			if ( t = RegSetValueEx( KeyHandle, DisabledAtStartRegValue, 0, REG_DWORD, (BYTE*) &HiddenAtStartup, sizeof( DWORD ) ) )
				rv = t;

		if ( HookList >= 0 )
			if ( t = RegSetValueEx( KeyHandle, HookListCmdRegValue, 0, REG_DWORD, (BYTE*) &HookList, sizeof( DWORD ) ) )
				rv = t;

		hotkeyval = RegGetDWord( KeyHandle, HotKeyRegValue, 0, UINT_MAX, UINT_MAX, NULL );

		if ( hotkeyval && !IsKnownKeyValue( hotkeyval ) )
			if ( t = RegSetValueEx( KeyHandle, HotKeyRegValue, 0, REG_DWORD, (BYTE*) Hotkey, sizeof( DWORD ) ) )
				rv = t;

		RegCloseKey( KeyHandle );
	}
	else
		rv = t;

	return rv;
}



LPCTSTR RootKeyName( int n )
{
	static const wchar_t RootName[2][5]		= { L"HKCU", L"HKLM" };


	if ( n == 1 )
		return RootName[0];
	else if ( n == 2 )
		return RootName[1];
	else
		return NULL;
}



struct _FindMyTabData {
	HWND MyTab;
	LPTSTR Title;
};



BOOL CALLBACK FindMyTabProc$( HWND hWnd, LPARAM lParam )
{
	wchar_t Title[SHORT_BUF_LEN]		= L"";

	_FindMyTabData *Data		= (_FindMyTabData*) lParam;


	GetWindowText( hWnd, Title, SHORT_BUF_LEN );

	if ( !wcscmp( Title, Data->Title ) ) {
		Data->MyTab = hWnd;
		return FALSE;
	}

	return TRUE;
}



HWND FindMyTab( HWND TakeCmdWindow, HWND ConsoleWindow )
{
	if ( !TakeCmdWindow || !ConsoleWindow )
		return NULL;


	wchar_t OldTitle[SHORT_BUF_LEN]		= L"";
	wchar_t TestTitle[SHORT_BUF_LEN]	= L"";

	_FindMyTabData Data;


	Data.MyTab		= NULL;
	Data.Title		= TestTitle;


	wsprintf( TestTitle, L"\u00ab%08X-%04X\u00bb", GetCurrentProcessId(), rand() );
	GetWindowText( ConsoleWindow, OldTitle, SHORT_BUF_LEN );
	SetWindowText( ConsoleWindow, TestTitle );

	for ( int i = 0; i < 200; i++ ) {
		Sleep( 10 );
		EnumChildWindows( TakeCmdWindow, FindMyTabProc$, (LPARAM) &Data );
		if ( Data.MyTab )
			break;
	}

	SetWindowText( ConsoleWindow, OldTitle );
	return Data.MyTab;
}



BOOL CALLBACK FindViewWindowProc$( HWND hWnd, LPARAM lParam )
{
	wchar_t Class[SHORT_BUF_LEN]		= L"";

	HWND *ViewWindow				= (HWND*) lParam;


	RealGetWindowClass( hWnd, Class, SHORT_BUF_LEN );

	if ( wcsncmp( Class, L"AfxFrameOrView", 14 ) )
		return TRUE;

	*ViewWindow = hWnd;
	return FALSE;
}



HWND FindViewWindow( HWND TabWindow )
{
	if ( !TabWindow )
		return NULL;


	HWND ViewWindow			= NULL;


	EnumChildWindows( TabWindow, FindViewWindowProc$, (LPARAM) &ViewWindow );
	return ViewWindow;
}



COLORREF DifferentColor( COLORREF Color )
{
	COLORREF Different		= 0x000000;


	if ( ( Color & 0x000080 ) == 0 )
		Different |= 0x0000ff;

	if ( ( Color & 0x008000 ) == 0 )
		Different |= 0x00ff00;

	if ( ( Color & 0x800000 ) == 0 )
		Different |= 0xff0000;

	return Different;
}



HWND FindTakeCmdWindow()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	HWND TakeCmd				= NULL;


	TakeCommandIPC( L"HWND", temp );

	if ( iswdigit( temp[0] ) ) {
		_set_errno( 0 );
		TakeCmd = (HWND) wcstoul( temp, NULL, 10 );
		if ( errno )
			TakeCmd = NULL;
	}

	return TakeCmd;
}



BOOL MarkInList( unsigned int Mark, unsigned int *AltMarks, unsigned int NumAltMarks )
{
	for ( unsigned int i = 0; i < NumAltMarks; i++ ) {
		if ( AltMarks[i] & RULER_FLAG_MARK_REPEATS ) {
			if ( ( Mark % ( AltMarks[i] & ~RULER_FLAG_MARK_REPEATS ) ) == 0 )
				return TRUE;
		}
		else if ( Mark == AltMarks[i] )
			return TRUE;
	}

	return FALSE;
}



int DrawRuler( HWND hWnd, _RulerWinData *Data )
{

#ifdef DEBUG_CODE

	if ( Debug )
		_cwprintf( L"   *  DrawRuler( 0x%08X, 0x%08I64X )  (%s ruler)\n", hWnd, (unsigned __int64) Data, RulerModeStr( Data->Mode ) );

#endif	//	DEBUG_CODE


	if ( Data->DesktopHack ) {
		ValidateRect( hWnd, NULL );
		return 0;
	}



	//		This is kludgey:  Sometimes the hidden console window causes an unending stream
	//		of WM_PAINT messages to be send to the ruler window.  (Because Take Command is
	//		repeatedly hiding and re-hiding the console window?)  Bad Things ensue.  The
	//		following routine watches for this situation.  If we see a long series of
	//		WM_PAINTs while the console window is hidden, we reassign the ruler window to
	//		the desktop window, breaking the cycle.

	if ( ( Data->Mode == RULER_MODE_CONSOLE ) && ( Data->DesktopHack == FALSE ) ) {
		if ( !IsWindowVisible( ConsoleWindow ) ) {
			Data->HiddenRedrawCount += 1;

#ifdef DEBUG_CODE

			if ( Debug )
				_cwprintf( L"   *  Console window is invisible (%u)\n", Data->HiddenRedrawCount );

#endif	//	DEBUG_CODE

			ShowWindow( hWnd, SW_SHOW );
			ValidateRect( hWnd, NULL );

			if ( Data->HiddenRedrawCount >= 10 ) {
				SetParent( hWnd, DesktopWindow );
				Data->DesktopHack = TRUE;
				Data->HiddenRedrawCount = 0;

#ifdef DEBUG_CODE

				if ( Debug )
					_cwprintf( L"   *  Set DesktopHack; GetParent( 0x%08X ) returns 0x%08X.\n", hWnd, GetParent( hWnd ) );

#endif	//	DEBUG_CODE

			}

			return 0;
		}
		else
			Data->HiddenRedrawCount = 0;
	}


	//		Looks like we really do need to draw the ruler :

	HDC Draw					= GetDC( hWnd );
	HBRUSH BackBrush			= NULL;
	HPEN Pen					= NULL;
	HPEN Pen2					= NULL;
	RECT Rect					= { 0, 0, 0, 0 };

	LONG RulerLeft				= 0;
	LONG RulerTop				= 0;
	LONG RulerWidth				= 0;
	LONG RulerHeight			= 0;

	HWND ViewWindow				= NULL;
	SCROLLBARINFO VScrollBar;
	SCROLLBARINFO HScrollBar;

	unsigned int ClientWidth	= 0;
	unsigned int ClientHeight	= 0;
	unsigned int CharWidth		= 0;
	unsigned int CharHeight		= 0;
	int t						= 0;
	int r						= 0;
	BOOL AltColorSelected		= FALSE;
	BOOL UseAltColor			= FALSE;
	COLORREF AltColor			= MAX_COLOR;



	//		Collect information on the console buffer and the underlying window :

	memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );
	GetConsoleScreenBufferInfo( ConsoleHandle, &BufferInfo );
	ViewportWidth		= BufferInfo.srWindow.Right - BufferInfo.srWindow.Left + 1;
	ViewportHeight		= BufferInfo.srWindow.Bottom - BufferInfo.srWindow.Top + 1;
	RulerColor			= RulerColorForBG[ ( BufferInfo.wAttributes & 0x00f0 ) >> 4 ];

	GetClientRect( Data->Parent, &ClientArea );
	ClientWidth		= ClientArea.right;
	ClientHeight	= ClientArea.bottom;

	CharWidth	= ( ClientWidth - Data->HMargin ) / ViewportWidth;
	CharHeight	= ( ClientHeight - Data->VMargin ) / ViewportHeight;


	//		If Take Command has scroll bars, they need to be deducted from
	//		the client area :

	if ( Data->Mode == RULER_MODE_TCMD ) {
		ViewWindow = FindViewWindow( Data->Parent );

		if ( ViewWindow ) {
			memset( &VScrollBar, 0x00, sizeof( SCROLLBARINFO ) );
			memset( &HScrollBar, 0x00, sizeof( SCROLLBARINFO ) );
			VScrollBar.cbSize	= sizeof( SCROLLBARINFO );
			HScrollBar.cbSize	= sizeof( SCROLLBARINFO );
			
			GetScrollBarInfo( ViewWindow, OBJID_VSCROLL, &VScrollBar );
			GetScrollBarInfo( ViewWindow, OBJID_HSCROLL, &HScrollBar );

			if ( ( VScrollBar.rgstate[0] & STATE_SYSTEM_INVISIBLE ) == 0 )
				ClientWidth -= VScrollBar.dxyLineButton;

			if ( ( HScrollBar.rgstate[0] & STATE_SYSTEM_INVISIBLE ) == 0 )
				ClientHeight = HScrollBar.dxyLineButton;
		}

		CharWidth	= ( ClientWidth - Data->HMargin ) / ViewportWidth;
		CharHeight	= ( ClientHeight - Data->HMargin ) / ViewportHeight;
	}

#ifdef DEBUG_CODE

	if ( Debug ) {
		_cwprintf( L"   *  Viewport:  %u x %u characters\n", ViewportWidth, ViewportHeight );
		_cwprintf( L"   *  Client width:  %u   Client height: %u pixels\n", ClientWidth, ClientHeight );
		_cwprintf( L"   *  Character size:  %u x %u pixels\n", CharWidth, CharHeight );
	}

#endif	//	DEBUG_CODE


	//		Position the ruler window where it belongs :

	switch ( Position ) {

	case RULER_POSITION_BOTTOM :

		RulerLeft	= 0;
		RulerTop	= ClientHeight - CharHeight - 1;
		RulerWidth	= ClientWidth;
		RulerHeight	= CharHeight;

		UseAltColor	= ( NumAltColumns > 0 );
		AltColor	= AltColumnsColor;
		break;

	case RULER_POSITION_LEFT :

		RulerLeft	= 0;
		RulerTop	= 0;
		RulerWidth	= CharWidth * 2;
		RulerHeight	= ClientHeight;

		if ( TopOffset ) {
			RulerTop	+= ( ( TopOffset * CharHeight ) + Data->VMargin );
			RulerHeight	-= ( ( TopOffset * CharHeight ) + Data->VMargin );
		}

		UseAltColor	= ( NumAltRows > 0 );
		AltColor	= AltRowsColor;
		break;

	case RULER_POSITION_RIGHT :

		RulerLeft	= ClientWidth - ( 2 * CharWidth ) - 1;
		RulerTop	= 0;
		RulerWidth	= CharWidth * 2;
		RulerHeight	= ClientHeight;

		if ( TopOffset ) {
			RulerTop	+= ( ( TopOffset * CharHeight ) + Data->VMargin );
			RulerHeight	-= ( ( TopOffset * CharHeight ) + Data->VMargin );
		}

		UseAltColor	= ( NumAltRows > 0 );
		AltColor	= AltRowsColor;
		break;

	default :

		RulerLeft	= 0;
		RulerTop	= 0;
		RulerWidth	= ClientWidth;
		RulerHeight	= CharHeight;

		if ( TopOffset )
			RulerTop = ( TopOffset * CharHeight ) + Data->VMargin;

		UseAltColor	= ( NumAltColumns > 0 );
		AltColor	= AltColumnsColor;
	}

	SetWindowPos( hWnd, HWND_TOP, RulerLeft, RulerTop, RulerWidth, RulerHeight, SWP_SHOWWINDOW );

#ifdef DEBUG_CODE

	if ( Debug )
		_cwprintf( L"   *  SetWindowPos( 0x%08X, HWND_TOP, %d, %d, %d, %d, SWP_SHOWWINDOW )\n", hWnd, RulerLeft, RulerTop, RulerWidth, RulerHeight );

#endif	//	DEBUG_CODE


	//		Set the transparency mode and prepare to draw :

	if ( TranspMode ) {
		SetLayeredWindowAttributes( hWnd, 0, (BYTE) Alpha, LWA_ALPHA );
		SetBkColor( Draw, BackColor );
		SetBkMode( Draw, OPAQUE );

		BackBrush = CreateSolidBrush( BackColor );
		Pen = CreatePen( PS_SOLID, PenWidth, HashColor );

		if ( AltColor == BackColor )
			UseAltColor = FALSE;

		if ( ( AltColor == HashColor ) && ( PenWidth == PenWidth2 ) )
			UseAltColor = FALSE;

		if ( UseAltColor )
			Pen2 = CreatePen( PS_SOLID, PenWidth2, AltColor );
	}
	else {
		COLORREF ColorKey = DifferentColor( RulerColor );

		if ( UseAltColor && ( ColorKey == AltColor ) )
			ColorKey ^= 0x030000;

		SetLayeredWindowAttributes( hWnd, ColorKey, (BYTE) Alpha, LWA_COLORKEY );
		SetBkColor( Draw, ColorKey );
		SetBkMode( Draw, OPAQUE );

		BackBrush = CreateSolidBrush( ColorKey );
		Pen = CreatePen( PS_SOLID, PenWidth, RulerColor );

		if ( ( AltColor == RulerColor ) && ( PenWidth == PenWidth2 ) )
			UseAltColor = FALSE;

		if ( UseAltColor )
			Pen2 = CreatePen( PS_SOLID, PenWidth2, AltColor );
	}

	SelectObject( Draw, BackBrush );
	SelectObject( Draw, Pen );


	//		Fill the ruler with the background color :

	Rect.right	= RulerWidth;
	Rect.bottom	= RulerHeight;

	FillRect( Draw, &Rect, BackBrush );

#ifdef DEBUG_CODE

	if ( Debug )
		_cwprintf( L"   *  FillRect( Draw, ( %u, %u ) - ( %u, %u ), BackBrush )\n", Rect.left, Rect.top, Rect.bottom, Rect.right );

#endif	//	DEBUG_CODE


	//		Draw the tick marks :

	if ( !Disabled ) {

		switch ( Position ) {

		case RULER_POSITION_LEFT :

			for ( unsigned int i = 1; i <= ViewportHeight; i++ ) {

				if ( UseAltColor ) {
					if ( MarkInList( i, AltRows, NumAltRows ) ) {
						if ( !AltColorSelected )
							SelectObject( Draw, Pen2 );

						AltColorSelected = TRUE;
					}
					else {
						if ( AltColorSelected )
							SelectObject( Draw, Pen );

						AltColorSelected = FALSE;
					}
				}

				if ( ( i % 10 ) == 0 )
					r = CharWidth * 7 / 4;
				else if ( ( i % 5 ) == 0 )
					r = CharWidth;
				else
					r = CharWidth / 2;

				t = ( i * CharHeight ) + Data->VMargin;

				MoveToEx( Draw, 0, t, NULL );
				LineTo( Draw, r, t );
			}

			break;

		case RULER_POSITION_RIGHT :

			for ( unsigned int i = 1; i <= ViewportHeight; i++ ) {

				if ( UseAltColor ) {
					if ( MarkInList( i, AltRows, NumAltRows ) ) {
						if ( !AltColorSelected )
							SelectObject( Draw, Pen2 );

						AltColorSelected = TRUE;
					}
					else {
						if ( AltColorSelected )
							SelectObject( Draw, Pen );

						AltColorSelected = FALSE;
					}
				}

				if ( ( i % 10 ) == 0 )
					r = CharWidth / 4;
				else if ( ( i % 5 ) == 0 )
					r = CharWidth;
				else
					r = CharWidth * 3 / 2;

				t = ( i * CharHeight ) + Data->VMargin;

				MoveToEx( Draw, RulerWidth, t, NULL );
				LineTo( Draw, r, t );
			}

			break;

		case RULER_POSITION_BOTTOM :

			for ( unsigned int i = 1; i <= ViewportWidth; i++ ) {

				if ( UseAltColor ) {
					if ( MarkInList( i, AltColumns, NumAltColumns ) ) {
						if ( !AltColorSelected )
							SelectObject( Draw, Pen2 );

						AltColorSelected = TRUE;
					}
					else {
						if ( AltColorSelected )
							SelectObject( Draw, Pen );

						AltColorSelected = FALSE;
					}
				}

				if ( ( i % 10 ) == 0 )
					t = CharHeight / 8;
				else if ( ( i % 5 ) == 0 )
					t = CharHeight / 2;
				else
					t = CharHeight * 3 / 4;

				r = i * CharWidth + Data->HMargin;

				MoveToEx( Draw, r, t, NULL );
				LineTo( Draw, r, RulerHeight );
			}

			break;

		default :

			for ( unsigned int i = 1; i <= ViewportWidth; i++ ) {

				if ( UseAltColor ) {
					if ( MarkInList( i, AltColumns, NumAltColumns ) ) {
						if ( !AltColorSelected )
							SelectObject( Draw, Pen2 );

						AltColorSelected = TRUE;
					}
					else {
						if ( AltColorSelected )
							SelectObject( Draw, Pen );

						AltColorSelected = FALSE;
					}
				}

				if ( ( i % 10 ) == 0 )
					t = CharHeight * 7 / 8;
				else if ( ( i % 5 ) == 0 )
					t = CharHeight / 2;
				else
					t = CharHeight / 4;

				r = i * CharWidth + Data->HMargin;

				MoveToEx( Draw, r, 0, NULL );
				LineTo( Draw, r, t );
			}
		}

	}


	//		Clean up and exit :

	ValidateRect( hWnd, NULL );
	DeleteObject( BackBrush );
	DeleteObject( Pen );
	DeleteObject( Pen2 );
	ReleaseDC( hWnd, Draw );

#ifdef DEBUG_CODE

	if ( Debug )
		_cwprintf( L"   *  DrawRuler() - normal exit (%s ruler)\n", RulerModeStr( Data->Mode ) );

#endif	//	 DEBUG_CODE

	if ( Disabled )
		ShowWindow( hWnd, SW_HIDE );

	return 0;
}



int RedrawRulers()
{
	int n		= 0;


	if ( ConsoleRuler ) {
		PostMessage( ConsoleRuler, WM_PAINT, 0, 0 );
		n++;
	}

	if ( TakeCmdRuler ) {
		PostMessage( TakeCmdRuler, WM_PAINT, 0, 0 );
		n++;
	}

	return n;
}



LRESULT CALLBACK RulerWinProc( HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam )
{
	_RulerWinData *Data		= (_RulerWinData*) GetWindowLongPtr( hWnd, GWLP_USERDATA );


	switch ( Msg ) {

	case WM_DESTROY :

#ifdef DEBUG_CODE

		if ( Debug )
			_cwprintf( L"   #  WM_DESTROY   hWnd = 0x%08X  %s\n", hWnd, RulerModeStr( Data ) );

#endif	//	DEBUG_CODE

		if ( Data->Mode == RULER_MODE_CONSOLE )
			ConsoleRuler = NULL;
		else if ( Data->Mode == RULER_MODE_TCMD )
			TakeCmdRuler = NULL;

		free( Data );
		PostQuitMessage( 0 );
		break;

	case WM_PAINT :

#ifdef DEBUG_CODE

		if ( Debug )
			_cwprintf( L"   #  WM_PAINT   hWnd = 0x%08X  %s\n", hWnd, RulerModeStr( Data ) );

#endif	//	DEBUG_CODE

		DrawRuler( hWnd, Data );
		return 0;

	default :

#ifdef DEBUG_CODE

		if ( Debug )
			_cwprintf( L"   #  Other window message:  %4u  0x%04X   hWnd = 0x%08X  %s\n", Msg, Msg, hWnd, RulerModeStr( Data ) );

#endif	//	DEBUG_CODE

		return DefWindowProc( hWnd, Msg, wParam, lParam );
	}

	return 0;
}



HWND CreateRulerWindow( unsigned int Mode, HWND Parent )
{
	wchar_t WinTitle[]			= L"Take Command Ruler";

	HWND hWnd					= NULL;
	_RulerWinData *Data			= NULL;



	//		Create a data block for variables associated with the window :

	Data = (_RulerWinData*) malloc( sizeof( _RulerWinData ) );
	if ( !Data ) {
		SetLastError( ERROR_NOT_ENOUGH_MEMORY );
		return NULL;
	}

	memset( Data, 0x00, sizeof( _RulerWinData ) );

	Data->Mode		= Mode;
	Data->Parent	= Parent;

	if ( Mode == RULER_MODE_TCMD ) {
		Data->HMargin		= TCmdHMargin;
		Data->VMargin		= TCmdVMargin;
	}


	//		Funny stuff here:  *Some* older versions of TCC bollix creating child windows of the
	//		console window.  (Not all windows messages are delivered to the child's window procedure,
	//		and CreateWindowEx() fails with the hysterically unhelpful ERROR_SUCCESS.)
	//
	//		So...  I'm creating the ruler window as a child of the desktop window, and then fostering
	//		it out to the desired parent window after it is successfully created.


	hWnd = CreateWindowEx( WS_EX_TOPMOST | WS_EX_LAYERED | WS_EX_TRANSPARENT | WS_EX_NOACTIVATE, RulerWindowClass, WinTitle, WS_CHILD, 0, 0, 100, 50, DesktopWindow, NULL, MyDLLHandle, NULL );
	if ( !hWnd ) {
		free( Data );
		return NULL;
	}

	SetWindowLongPtr( hWnd, GWLP_USERDATA, (LONG_PTR) Data );
	SetParent( hWnd, Parent );
	SetLayeredWindowAttributes( hWnd, 0, RULER_DEFAULT_ALPHA, LWA_ALPHA );
	return hWnd;
}



void CALLBACK TimerProc( HWND hWnd, UINT Msg, UINT_PTR TimerID, DWORD UpTime )
{
	CONSOLE_SCREEN_BUFFER_INFO BufInfo;

	BOOL Visible				= IsWindowVisible( ConsoleWindow );
	_RulerWinData *Data			= NULL;
	HMONITOR ConsoleMonitor		= NULL;
	RECT ClientAreaNow			= { 0, 0, 0, 0 };
	static unsigned int k		= 0;



	if ( ConsoleRuler )
		Data	= (_RulerWinData*) GetWindowLongPtr( ConsoleRuler, GWLP_USERDATA );


	//		Handle delayed recheck :

	if ( DelayedRedraw )
		if ( --DelayedRedraw == 0 )
			RedrawRulers();


	//		Check whether console window visibility has changed :

	if ( Visible != LastVisible ) {

#ifdef DEBUG_CODE

		if ( Debug )
			_cputws( L"   +  Console visibility has changed!\n" );

#endif	//	DEBUG_CODE

		LastVisible = Visible;
		DelayedRedraw = 3;

		if ( ConsoleRuler ) {
			if ( Visible ) {

#ifdef DEBUG_CODE

				if ( Debug )
					_cputws( L"   +  Console window just became visible.\n" );

#endif	//	DEBUG_CODE

				if ( Data ) {
					if ( Data->DesktopHack ) {
						Data->DesktopHack = FALSE;
						Data->HiddenRedrawCount = 0;
						SetParent( ConsoleRuler, Data->Parent );
						SetWindowPos( ConsoleRuler, HWND_TOP, 0, 0, 50, 50, SWP_SHOWWINDOW );

#ifdef DEBUG_CODE

						if ( Debug )
							_cwprintf( L"   +  Re-attached console ruler 0x%08X to console window 0x%08X.\n", ConsoleRuler, Data->Parent );

#endif	//	DEBUG_CODE

					}
				}
			}
			else {
				TakeCmdWindow = FindTakeCmdWindow();

				if ( TakeCmdWindow )
					MyTab = FindMyTab( TakeCmdWindow, ConsoleWindow );
				else
					MyTab = NULL;

				if ( MyTab && !TakeCmdRuler ) {
					TakeCmdRuler = CreateRulerWindow( RULER_MODE_TCMD, MyTab );

#ifdef DEBUG_CODE

					if ( Debug && TakeCmdRuler )
						_cwprintf( L"   +  Created new TakeCmdRuler = 0x%08X\n", TakeCmdRuler );

#endif	//	DEBUG_CODE

				}
			}
		}
	}


	//		Check whether the viewport size has changed :

	if ( !DelayedRedraw ) {
		memset( &BufInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );
		GetConsoleScreenBufferInfo( ConsoleHandle, &BufInfo );

		if ( ( ( BufInfo.srWindow.Right - BufInfo.srWindow.Left + 1 ) != ViewportWidth ) || ( ( BufInfo.srWindow.Bottom - BufInfo.srWindow.Top + 1 ) != ViewportHeight  ) )
			DelayedRedraw = 1;

		if ( ( BufInfo.wAttributes & 0x00f0 ) != ( BufferInfo.wAttributes & 0x00f0 ) )
			DelayedRedraw = 1;
	}


	//		Stuff that happens every eighth interrupt :

	k = ( k + 1 ) % 8;

	if ( !k ) {
		
		//		Check whether the console window has been moved to a different monitor :

		if ( Visible && !DelayedRedraw ) {
			ConsoleMonitor = MonitorFromWindow( ConsoleWindow, MONITOR_DEFAULTTONEAREST );

			if ( ConsoleMonitor != LastConsoleMonitor ) {
				if ( LastConsoleMonitor ) {

#ifdef DEBUG_CODE

					if ( Debug )
						_cputws( L"   +  Console window moved to a different monitor!\n" );

#endif	//	DEBUG_CODE

					DelayedRedraw = 1;
				}

				LastConsoleMonitor = ConsoleMonitor;
			}
		}

		//		Check whether the console window has changed size :

		if ( Visible && !DelayedRedraw ) {
			GetClientRect( ConsoleWindow, &ClientAreaNow );

			if ( ( ClientAreaNow.right != ClientArea.right ) || ( ClientAreaNow.bottom != ClientArea.bottom ) )
				DelayedRedraw = 1;
		}


		if ( !Visible && MyTab && !TakeCmdRuler ) {		//	check for tab detached and hidden :
			if ( !IsWindow( MyTab ) ) {
				MyTab = NULL;
				TakeCmdWindow = NULL;
			}
		}
		else if ( !Visible && !TakeCmdWindow ) {		//	check for invisible console attached :
			TakeCmdWindow = FindTakeCmdWindow();

			if ( TakeCmdWindow )
				MyTab = FindMyTab( TakeCmdWindow, ConsoleWindow );
			else
				MyTab = NULL;

			if ( MyTab ) {
				TakeCmdRuler = CreateRulerWindow( RULER_MODE_TCMD, MyTab );

#ifdef DEBUG_CODE

				if ( Debug && TakeCmdRuler )
					_cwprintf( L"   +  Created new TakeCmdRuler = 0x%08X\n", TakeCmdRuler );

#endif	//	DEBUG_CODE

			}
		}

		if ( Data )
			Data->HiddenRedrawCount = 0;
	}

	return;
}



int DisableOneFeature( LPTSTR FeaturesList, LPTSTR word )
{
	if ( !FeaturesList || !word )
		return -1;

	if ( ( FeaturesList[0] == '\0' ) || ( word[0] == '\0' ) )
		return -1;


	size_t i = 0, j = 0, base = 0;
	wchar_t ch = '\0', ch2 = '\0';
	BOOL done = FALSE, match = FALSE;


	do {
		// Compare word in 'word' against word at 'base' :

		i = 0;
		match = TRUE;

		do {
			ch = towupper( FeaturesList[base + i] );
			ch2 = ch;

			if ( ( ch2 == ',' ) || ( ch2 == ';' ) )
				ch2 = '\0';

			match &= ( ch2 == towupper( word[i] ) );

			i++;
		}
		while ( ch2 );

		if ( match ) {
			// disable the feature:

			if ( ch == '\0' )
				FeaturesList[base] = ch;
			else {
				j = 0;

				do {
					ch = FeaturesList[base + i + j];
					FeaturesList[base + j] = ch;
					j++;
				}
				while ( ch );
			}
		}
		else
			base += i;

	}
	while ( ch );


	i = wcslen( FeaturesList );

	if ( i )
		if ( ( FeaturesList[i - 1] == ',' ) || ( FeaturesList[i - 1] == ';' ) )
			FeaturesList[i - 1] = '\0';

	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/ruler.html";
	piInfo.pszDescription	= (LPTSTR) L"On-screen ruler for Take Command";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN]		= L"";

	HMODULE User32Handle				= GetModuleHandle( L"user32.dll" );

	typedef BOOL ( WINAPI _SetProcDpiAware ) ( void );
	_SetProcDpiAware *SetProcDpiAware	= NULL;



	//		Dynamically link to and call SetProcessDPIAware() :

	if ( User32Handle ) {
		SetProcDpiAware = (_SetProcDpiAware*) GetProcAddress( User32Handle, "SetProcessDPIAware" );

		if ( SetProcDpiAware )
			SetProcDpiAware();
	}


	//		Get info about TCC and Take Command :

	GetTCCVer();

	if ( TCCVerMajor < 13 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Take Command version 13 or later is required!\r\n" );
		goto Fail;
	}

	if ( TCCVerMajor >= 24 )
		GetTCmdMargins();
	else
		TCmdHMargin = TCmdVMargin = 0;


	TakeCmdWindow	= FindTakeCmdWindow();
	ConsoleWindow	= GetConsoleWindow();
	LastVisible		= IsWindowVisible( ConsoleWindow );
	DesktopWindow	= GetDesktopWindow();


	if ( TakeCmdWindow )
		MyTab = FindMyTab( TakeCmdWindow, ConsoleWindow );
	else
		MyTab = NULL;


	//		Dynamic link to Kernel32 functions :

	Kernel32Handle = GetModuleHandle( L"Kernel32.dll" );
	if ( Kernel32Handle ) 
		GetCSBInfoExProc = (GET_CSB_INFO_EX_PROC) GetProcAddress( Kernel32Handle, "GetConsoleScreenBufferInfoEx" );


	//		Default plugin settings :

	InitHotkeys();
	RestorePluginDefaults();
	RestoreOptions();


	//		Disable LIST command hook, if desired :

	if ( !HookListCmd )
		DisableOneFeature( PluginFeatures, L"list" );


	//		Get console handle :

	ConsoleHandle = CreateFile( L"conout$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );

	if ( ConsoleHandle == INVALID_HANDLE_VALUE ) {
		GetLastErrorText( shortbuf );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Failed to open console buffer:\r\n" );
		Qprintf( ERRHANDLE, L"  %s\r\n", shortbuf );
		goto Fail;
	}


	//		Register the window class :

	wsprintf( RulerWindowClass, L"Take Command Ruler %04X", GetCurrentProcessId() );

	WNDCLASSEX RulerWinClass;

	memset( &RulerWinClass, 0x00, sizeof( WNDCLASSEX ) );
	RulerWinClass.cbSize			= sizeof( WNDCLASSEX );
	RulerWinClass.lpfnWndProc		= RulerWinProc;
	RulerWinClass.hInstance			= MyDLLHandle;
	RulerWinClass.lpszClassName		= RulerWindowClass;


	RulerWindowAtom = RegisterClassEx( &RulerWinClass );

	if ( !RulerWindowAtom ) {
		GetLastErrorText( shortbuf );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Failed to register ruler window class:\r\n" );
		Qprintf( ERRHANDLE, L"  %s\r\n", shortbuf );
		goto Fail;
	}


	if ( DisabledAtStartup )
		Disabled = TRUE;


	ConsoleRuler = CreateRulerWindow( RULER_MODE_CONSOLE, ConsoleWindow );

	if ( !ConsoleRuler ) {
		GetLastErrorText( shortbuf );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Failed to create console ruler window:\r\n" );
		Qprintf( ERRHANDLE, L"  %s\r\n", shortbuf );
		goto Fail;
	}


	if ( MyTab )
		TakeCmdRuler = CreateRulerWindow( RULER_MODE_TCMD, MyTab );


	TimerID = SetTimer( NULL, RULER_DEFAULT_TIMER_ID, RULER_TIMER_TIMEOUT, TimerProc );


	wcscpy_s( shortbuf, SHORT_BUF_LEN, L"%[_transient]%[_pipe]%[noloadmsg]" );
	ExpandVariables( shortbuf, 1 );

	if ( wcscmp( shortbuf, L"00" ) == 0 ) {
		shortbuf[0] = '\0';
		GetKeyName( Hotkey[0], shortbuf, TRUE );
		Printf( L"%s v%u.%u.%u loaded; hotkey is %s.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, shortbuf );
	}

#ifdef DEBUG_CODE

	if ( Debug ) {
		_cwprintf( L"   :  TakeCmdWindow = 0x%08X   ConsoleWindow = 0x%08X   DesktopWindow = 0x%08X\n", TakeCmdWindow, ConsoleWindow, DesktopWindow );
		_cwprintf( L"   :  ConsoleRuler = 0x%08X   TakeCmdRuler = 0x%08X\n", ConsoleRuler, TakeCmdRuler );
		_cwprintf( L"   :  TimerID = 0x%08X\n", TimerID );
	}

#endif	//	DEBUG_CODE


	return 0;


Fail:

	if ( TimerID )
		KillTimer( NULL, TimerID );

	if ( ConsoleRuler )
		DestroyWindow( ConsoleRuler );

	if ( TakeCmdRuler )
		DestroyWindow( TakeCmdRuler );

	if ( RulerWindowAtom )
		UnregisterClass( RulerWindowClass, MyDLLHandle );

	CloseHandle( ConsoleHandle );

	Qprintf( ERRHANDLE, L"Plugin not loaded.\r\n" );
	return 1;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	if ( TimerID )
		KillTimer( NULL, TimerID );

	if ( ConsoleRuler )
		DestroyWindow( ConsoleRuler );

	if ( TakeCmdRuler )
		DestroyWindow( TakeCmdRuler );

	if ( RulerWindowAtom )
		UnregisterClass( RulerWindowClass, MyDLLHandle );

	CloseHandle( ConsoleHandle );

	return 0;
}



LPCTSTR OrdinalString( unsigned int Value )
{
	static wchar_t temp[64]				= L"";
	static const wchar_t ord[4][3]		= { L"th", L"st", L"nd", L"rd" };
	unsigned int n						= Value % 100;
	unsigned int i						= 0;


	if ( ( n >= 10 ) && ( n < 20 ) )
		i = 0;
	else {
		i = n % 10;
		if ( i > 3 )
			i = 0;
	}

	wsprintf( temp, L"%u%s", Value, ord[i] );
	return temp;
}



int GetCurrentConsoleColors()
{
	int rv				= 0;
	
	CONSOLE_SCREEN_BUFFER_INFOEX BufferInfo;


	for ( int i = 0; i < 16; i++ )
		CurrentConsoleColors[i] = DefaultConsoleColors[i];


	if ( GetCSBInfoExProc ) {
		memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFOEX ) );
		BufferInfo.cbSize = sizeof( CONSOLE_SCREEN_BUFFER_INFOEX );

		GetCSBInfoExProc( ConsoleHandle, &BufferInfo );

		if ( BufferInfo.dwSize.X && BufferInfo.dwSize.Y ) {
			for ( int i = 0; i < 16; i++ )
				CurrentConsoleColors[i] = BufferInfo.ColorTable[i];

			rv = 1;
		}
	}

	return rv;
}



//		Returns a value from 0 (darkest) to 1000 (brightest).

int ColorRefLuminance( COLORREF Color )
{
	if ( Color > MAX_COLOR )
		return -1;


	return ( ( 2126 * GetRValue( Color ) ) + ( 7152 * GetGValue( Color ) ) + ( 722 * GetBValue( Color ) ) ) / 255;
}



int ShowCaretHint( LPCTSTR Arg, const wchar_t *Caret )
{
	if ( !Arg )
		return -1;


	wchar_t temp[MAX_ARG_LEN + 1]		= L"";
	size_t CaretAt						= Caret - Arg;


	for ( size_t i = 0; i < CaretAt; i++ )
		temp[i] = 0x20;

	temp[CaretAt + 0]	= '\0';

	Qprintf( ERRHANDLE, L"    %s\r\n", Arg );
	Qprintf( ERRHANDLE, L"    %s\u2191\r\n", temp );
	return 0;
}



int ParseConsoleColorIndex( LPTSTR String, LPTSTR Arg )
{
	if ( !String )
		return -1;


	wchar_t *ch			= String;
	unsigned int n		= 0;


	if ( *ch == '$' )
		ch++;

	if ( ParseInt( ch, &n, NULL ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad console color index: \"%s\"\r\n", String );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	if ( n >= 16 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Console color index must be 0 - 15: %u\r\n", n );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	return CurrentConsoleColors[n];
}



int ParseRGBColorRef( LPTSTR String, LPTSTR Arg )
{
	if ( !String )
		return -1;


	wchar_t *ch			= String;
	wchar_t *sep1		= NULL;

	unsigned int red	= 0;
	unsigned int green	= 0;
	unsigned int blue	= 0;


	if ( ( *ch == ',' ) || ( *ch == '\0' ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing red value\r\n" );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	if ( ParseInt( ch, &red, &sep1 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad red value:\r\n" );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	if ( red > 0xff ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Red value must be 0 - 255: %u\r\n", red );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	if ( *sep1 != ',' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Expected a comma:\r\n" );
		ShowCaretHint( Arg, sep1 );
		return -1;
	}

	ch = sep1 + 1;

	if ( ( *ch == ',' ) || ( *ch == '\0' ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing green value\r\n" );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	if ( ParseInt( ch, &green, &sep1 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad green value:\r\n" );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	if ( green > 0xff ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Green value must be 0 - 255: %u\r\n", green );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	if ( *sep1 != ',' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Expected a comma:\r\n" );
		ShowCaretHint( Arg, sep1 );
		return -1;
	}

	ch = sep1 + 1;

	if ( ( *ch == ',' ) || ( *ch == '\0' ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing blue value\r\n" );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	if ( ParseInt( ch, &blue, &sep1 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad blue value:\r\n" );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	if ( blue > 0xff ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Blue value must be 0 - 255: %u\r\n", blue );
		ShowCaretHint( Arg, ch );
		return -1;
	}

	if ( *sep1 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", sep1 );
		ShowCaretHint( Arg, sep1 );
		return -1;
	}
	
	return ( blue << 16 ) | ( green << 8 ) | red;
}



int ParseColor( LPTSTR String, LPTSTR CompleteArg )
{
	if ( !String )
		return -1;


	wchar_t *ch			= String;
	int ColorValue		= 0;


	if ( *ch == '/' ) {
		ch++;
		if ( iswalpha( *ch ) ) {
			ch++;
			if ( IsColonOrEquals( *ch ) )
				ch++;
		}
	}


	if ( *ch == '$' )
		return ParseConsoleColorIndex( ch, CompleteArg );

	if ( wcschr( ch, ',' ) )
		return ParseRGBColorRef( ch, CompleteArg );

	ColorValue = W3CColorValue( ch, TRUE );

	if ( ColorValue < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad color name: \"%s\"\r\n", String );
		ShowCaretHint( CompleteArg, ch );
		return -1;
	}

	return RGB2ColorRef( ColorValue );
}



int __cdecl AltMarksCmp$( const void *a1, const void *a2 )
{
	const unsigned int *v1	= (const unsigned int*) a1;
	const unsigned int *v2	= (const unsigned int*) a2;

	if ( *v1 < *v2 )
		return -1;
	else if ( *v1 > *v2 )
		return 1;
	else
		return 0;
}



int ParseAltMarks( LPTSTR inString, unsigned int *AltMarks, unsigned int *NumAltMarks, COLORREF *AltColor, LPCTSTR MarkType )
{
	if ( !inString )
		return -1;


	wchar_t MarkTypeIC[128]			= L"";

	wchar_t *ch						= inString;
	wchar_t *sep1					= NULL;
	unsigned int t					= 0;
	BOOL Found						= FALSE;
	int NewAltColor					= 0;
	unsigned int NewNumAltMarks		= 0;

	unsigned int NewAltMarks[RULER_MAX_ALT_MARKS];



	wcscpy_s( MarkTypeIC,128, MarkType );
	MarkTypeIC[0] = towupper( MarkTypeIC[0] );


	if ( *ch == '/' ) {
		ch++;

		while ( iswalpha( *ch ) )
			ch++;

		if ( IsColonOrEquals( *ch ) )
			ch++;
	}


	if ( *ch == '\0' ) {
		*NumAltMarks = 0;
		*AltColor = MAX_COLOR;
		return 0;
	}


	while ( *ch ) {

		if ( ( *ch == '*' ) || ( *ch == 0xd7 ) ) {
			if ( ParseInt( ch + 1, &t, &sep1 ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad %s repeat value:\r\n", MarkType );
				ShowCaretHint( inString, ch + 1 );
				return 1;
			}

			if ( ( t < 2 ) || ( t > RULER_MAX_COLUMN_REPEAT ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"%s repeat value must be 2 - %u: %u\r\n", MarkTypeIC, RULER_MAX_COLUMN_REPEAT, t );
				ShowCaretHint( inString, ch + 1 );
				return 1;
			}

			t |= RULER_FLAG_MARK_REPEATS;
		}
		else {
			if ( ParseInt( ch, &t, &sep1 ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad %s value:\r\n", MarkType );
				ShowCaretHint( inString, ch );
				return 1;
			}

			if ( ( t < 1 ) || ( t > RULER_MAX_COLUMN ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"%s must be 1 - %u: %u\r\n", MarkTypeIC, RULER_MAX_COLUMN, t );
				ShowCaretHint( inString, ch );
				return 1;
			}
		}


		Found = FALSE;

		for ( unsigned int j = 0; j < NewNumAltMarks; j++ )
			if ( t == NewAltMarks[j] )
				Found = TRUE;

		if ( !Found ) {
			if ( NewNumAltMarks >= RULER_MAX_ALT_MARKS ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Too many alt marks (%u max)\r\n", RULER_MAX_ALT_MARKS );
				ShowCaretHint( inString, ch );
				return 1;
			}

			NewAltMarks[NewNumAltMarks++] = t;
		}


		if ( IsColonOrEquals( *sep1 ) )
			break;

		if ( *sep1 != ',' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Expected either a comma or an equals sign:\r\n" );
			ShowCaretHint( inString, sep1 );
			return 1;
		}

		ch = sep1 + 1;
	}

	ch = sep1 + 1;

	if ( *ch == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Expected a color:\r\n" );
		ShowCaretHint( inString, ch );
		return 1;
	}

	NewAltColor = ParseColor( ch, inString );
	if ( NewAltColor < 0 )
		return 1;


	if ( NewNumAltMarks > 1 )
		qsort( NewAltMarks, NewNumAltMarks, sizeof( unsigned int ), AltMarksCmp$ );


	*NumAltMarks	= 0;

	for ( unsigned int i = 0; i < NewNumAltMarks; i++ )
		AltMarks[i] = NewAltMarks[i];

	*AltColor		= NewAltColor;
	*NumAltMarks	= NewNumAltMarks;
	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	if ( !HelpText )
		return 1;

	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI ruler( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t shortbuf[SHORT_BUF_LEN]			= L"";

	int SaveKey					= 0;
	BOOL Restore				= FALSE;
	BOOL PluginDefaults			= FALSE;
	BOOL ListKeys				= FALSE;
	BOOL RedrawNeeded			= FALSE;
	BOOL Quiet					= FALSE;

	int NewHotkey[RULER_NUM_HOTKEYS];
	int NewHotkeyMode[RULER_NUM_HOTKEYS];
	int NewBackColor			= -1;
	int NewHashColor			= -1;
	int NewAlpha				= -1;
	int NewPosition				= -1;
	int NewTopOffset			= -1;
	int NewWidth				= -1;
	int NewWidth2				= -1;
	int NewStartDisabled		= -1;
	int NewTranspMode			= -1;
	int NewShow					= -1;
	int NewHookListCmd			= -1;

	int trv						= 0;
	int rv						= 0;

	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;

	int NewRulerColor[16];
	unsigned int ConsoleBG		= 0;



	memset( &ConsoleInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );
	GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo );
	ConsoleBG = ( ConsoleInfo.wAttributes & 0x00f0 ) >> 4;


	GetCurrentConsoleColors();


	for ( int i = 0; i < RULER_NUM_HOTKEYS; i++ )
		NewHotkey[i] = NewHotkeyMode[i] = -1;

	for ( int i = 0; i < 16; i++ )
		NewRulerColor[i] = -1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_FORCE_UPPERCASE );


	for ( int i = 0; i < numargs; i++ ) {

		if ( !wcscmp( arg[i], L"/?" ) )
			return ShowCmdHelp( RulerHelpText, FALSE );
		else if ( !wcscmp( arg[i], L"/??" ) )
			return ShowCmdHelp( RulerHelpText, TRUE );

#ifdef DEBUG_CODE

		else if ( !wcscmp( arg[i], L"/$" ) )
			Debug = !Debug;

#endif	//	DEBUG_CODE

		else if ( !wcsncmp( arg[i], L"/A", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( arg[i][j] == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing alpha value\r\n" );
				return 1;
			}

			if ( ParseInt( arg[i] + j, (unsigned int*) &NewAlpha, NULL ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad alpha value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( ( NewAlpha < RULER_ALPHA_MIN ) || ( NewAlpha > RULER_ALPHA_MAX ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Alpha must be %u - %u: %u\r\n", RULER_ALPHA_MIN, RULER_ALPHA_MAX, NewAlpha );
				return 1;
			}

			RedrawNeeded = TRUE;
		}
		else if ( !wcsncmp( arg[i], L"/B", 2 ) ) {
			NewBackColor = ParseColor( arg[i], arg[i] );
			RedrawNeeded = TRUE;

			if ( NewBackColor < 0 )
				return 1;
		}
		else if ( !wcsncmp( arg[i], L"/C", 2 ) ) {
			NewRulerColor[ConsoleBG] = ParseColor( arg[i], arg[i] );
			RedrawNeeded = TRUE;

			if ( NewRulerColor[ConsoleBG] < 0 )
				return 1;
		}
		else if ( !wcscmp( arg[i], L"/D" ) )
			PluginDefaults = TRUE, RedrawNeeded = TRUE;
		else if ( !wcsncmp( arg[i], L"/E", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( arg[i][j] == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing show value\r\n" );
				return 1;
			}

			if ( arg[i][j] == 'N' )
				NewShow = 0;
			else if ( arg[i][j] == 'Y' )
				NewShow = 1;
			else if ( ParseInt( arg[i] + j, (unsigned int*) &NewShow, NULL ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad show value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( ( NewShow < 0 ) || ( NewShow > 1 ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Show must be 0 or 1: %u\r\n", NewShow );
				return 1;
			}

			RedrawNeeded = TRUE;
		}
		else if ( !wcsncmp( arg[i], L"/H", 2 ) ) {
			NewHashColor = ParseColor( arg[i], arg[i] );
			RedrawNeeded = TRUE;

			if ( NewHashColor < 0 )
				return 1;
		}
		else if ( !wcsncmp( arg[i], L"/I", 2 ) ) {
			if ( ParseAltMarks( arg[i], AltColumns, &NumAltColumns, &AltColumnsColor, L"column" ) )
				return 1;

			RedrawNeeded = TRUE;
		}
		else if ( !wcsncmp( arg[i], L"/J", 2 ) ) {
			if ( ParseAltMarks( arg[i], AltRows, &NumAltRows, &AltRowsColor, L"row" ) )
				return 1;

			RedrawNeeded = TRUE;
		}
		else if ( !wcsncmp( arg[i], L"/K", 2 ) ) {
			int j = 2;
			unsigned int n = 0;

			if ( ( arg[i][j] >= '1' ) && ( arg[i][j] <= ( '0' + RULER_NUM_HOTKEYS ) ) ) {
				n = arg[i][j] - '1';
				j++;
			}

			if ( IsColonOrEquals( arg[i][j] ) )
				j++;
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Expected colon or digit 1 - %u: \"%s\"\r\n", RULER_NUM_HOTKEYS, arg[i] );
				ShowCaretHint( arg[i], arg[i] + j );
				return 1;
			}

			if ( !wcscmp( arg[i] + j, L"NONE" ) )
				NewHotkey[n] = 0;
			else {
				NewHotkey[n] = ParseKeyName( arg[i] + j );

				if ( NewHotkey[n] == 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown key name: \"%s\"\r\n", arg[i] );
					return 1;
				}
				else if ( NewHotkey[n] == -1 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Invalid hotkey: \"%s\"\r\n", arg[i] );
					return 1;
				}
				else if ( NewHotkey[n] == -2 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Key not allowed: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
		}
		else if ( !wcscmp( arg[i], L"/L" ) )
			ListKeys = TRUE;
		else if ( !wcsncmp( arg[i], L"/M", 2 ) ) {
			int j = 2;
			unsigned int n = 0;

			if ( ( arg[i][j] >= '1' ) && ( arg[i][j] <= ( '0' + RULER_NUM_HOTKEYS ) ) ) {
				n = arg[i][j] - '1';
				j++;
			}

			if ( IsColonOrEquals( arg[i][j] ) )
				j++;
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Expected colon or digit 1 - %u: \"%s\"\r\n", RULER_NUM_HOTKEYS, arg[i] );
				ShowCaretHint( arg[i], arg[i] + j );
				return 1;
			}

			if ( arg[i][j] == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing hotkey mode value\r\n" );
				return 1;
			}

			if ( ParseInt( arg[i] + j, (unsigned int*) ( NewHotkeyMode + n ), NULL ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad hotkey mode value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( ( NewHotkeyMode[n] < 0 ) || ( NewHotkeyMode[n] > RULER_HOTKEY_MODE_MAX ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"New hotkey mode must be 0 - %u: %u\r\n", RULER_HOTKEY_MODE_MAX, NewHotkeyMode[n] );
				return 1;
			}
		}
		else if ( !wcsncmp( arg[i], L"/O", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( arg[i][j] == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing offset value\r\n" );
				return 1;
			}

			if ( ParseInt( arg[i] + j, (unsigned int*) &NewTopOffset, NULL ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad offset value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( ( NewTopOffset < 0 ) || ( NewTopOffset > 5 ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Offset must be 0 - 5: %u\r\n", NewTopOffset );
				return 1;
			}

			RedrawNeeded = TRUE;

			if ( numargs == 1 )
				Quiet = TRUE;
		}
		else if ( !wcsncmp( arg[i], L"/P", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( arg[i][j] == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing position value\r\n" );
				return 1;
			}

			if ( arg[i][j] == 'T' )
				NewPosition = 0;
			else if ( arg[i][j] == 'B' )
				NewPosition = 1;
			else if ( arg[i][j] == 'L' )
				NewPosition = 2;
			else if ( arg[i][j] == 'R' )
				NewPosition = 3;
			else if ( ParseInt( arg[i] + j, (unsigned int*) &NewPosition, NULL ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad position value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( ( NewPosition < 0 ) || ( NewPosition > 4 ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Position must be 0 - 4: %u\r\n", NewPosition );
				return 1;
			}

			RedrawNeeded = TRUE;
		}
		else if ( !wcscmp( arg[i], L"/Q" ) )
			Quiet = TRUE;
		else if ( !wcscmp( arg[i], L"/R" ) )
			Restore = TRUE, RedrawNeeded = TRUE;
		else if ( !wcscmp( arg[i], L"/S" ) )
			SaveKey = 1;
		else if ( !wcscmp( arg[i], L"/S!" ) )
			SaveKey = 2;
		else if ( !wcsncmp( arg[i], L"/U", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( arg[i][j] == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing hook-list value\r\n" );
				return 1;
			}

			if ( arg[i][j] == 'N' )
				NewHookListCmd = FALSE;
			else if ( arg[i][j] == 'Y' )
				NewHookListCmd = TRUE;
			else if ( ParseInt( arg[i] + j, (unsigned int*) &NewHookListCmd, NULL ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad hook-list value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( ( NewHookListCmd < 0 ) || ( NewHookListCmd > 1 ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"New hook-list value must be 0 or 1: %u\r\n", NewHookListCmd );
				return 1;
			}
		}
		else if ( !wcsncmp( arg[i], L"/W", 2 ) ) {
			wchar_t *ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			if ( *ch == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing width value\r\n" );
				return 1;
			}

			if ( ParseInt( ch, (unsigned int*) &NewWidth, &ch ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad width value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( ( NewWidth < 0 ) || ( NewWidth > RULER_PEN_WIDTH_MAX ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Width must be 0 - %u: %u\r\n", RULER_PEN_WIDTH_MAX, NewWidth );
				return 1;
			}

			if ( *ch == '\0' )
				NewWidth2 = NewWidth;
			else if ( ( *ch == ',' ) || ( *ch == ';' ) ) {
				ch++;

				if ( *ch == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing width2 value\r\n" );
					return 1;
				}

				if ( ParseInt( ch, (unsigned int*) &NewWidth2, NULL ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad width2 value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ( NewWidth2 < 0 ) || ( NewWidth2 > RULER_PEN_WIDTH_MAX ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Width2 must be 0 - %u: %u\r\n", RULER_PEN_WIDTH_MAX, NewWidth2 );
					return 1;
				}
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Expected a comma:\r\n" );
				ShowCaretHint( arg[i], ch );
				return 1;
			}

			RedrawNeeded = TRUE;
		}
		else if ( !wcsncmp( arg[i], L"/X", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( arg[i][j] == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing transpmode value\r\n" );
				return 1;
			}

			if ( arg[i][j] == 'C' )
				NewTranspMode = 0;
			else if ( arg[i][j] == 'A' )
				NewTranspMode = 1;
			else if ( ParseInt( arg[i] + j, (unsigned int*) &NewTranspMode, NULL ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad transpmode value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( ( NewTranspMode < 0 ) || ( NewTranspMode > 1 ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Transpmode must be 0 or 1: %u\r\n", NewTranspMode );
				return 1;
			}

			RedrawNeeded = TRUE;
		}
		else if ( !wcsncmp( arg[i], L"/Z", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( arg[i][j] == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing start-hidden value\r\n" );
				return 1;
			}

			if ( arg[i][j] == 'N' )
				NewStartDisabled = FALSE;
			else if ( arg[i][j] == 'Y' )
				NewStartDisabled = TRUE;
			else if ( ParseInt( arg[i] + j, (unsigned int*) &NewStartDisabled, NULL ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad start-hidden value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( ( NewStartDisabled < 0 ) || ( NewStartDisabled > 1 ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"New start-hidden value must be 0 or 1: %u\r\n", NewStartDisabled );
				return 1;
			}
		}
		else if ( arg[i][0] == '/' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			return 1;
		}
		else {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
			return 1;
		}
	}



	if ( ( NewBackColor != -1 ) && ( NewHashColor == -1 ) ) {
		if ( ColorRefLuminance( NewBackColor ) >= 5000 )
			NewHashColor = 0x000000;
		else
			NewHashColor = 0xffffff;
	}



	//		List supported keys if desired :

	if ( ListKeys ) {
		if ( SaveKey || Restore || PluginDefaults ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t combine /L with other options\r\n" );
			return 1;
		}

		ListLegalKeys();
		return 0;
	}


	//		Restore settings if desired :

	if ( PluginDefaults && Restore ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t combine /D and /R\r\n" );
		return 1;
	}

	if ( Restore ) {
		trv = RestoreOptions();

		if ( !Quiet ) {
			if ( trv )
				Printf( L"Settings restored from the registry (%s\\%s).\r\n", RootKeyName( trv ), PluginRegKey );
			else
				Printf( L"Settings not found in the registry.\r\n" );
		}
	}
	else if ( PluginDefaults ) {
		RestorePluginDefaults();

		if ( !Quiet )
			Printf( L"Default plugin settings restored.\r\n" );
	}


	//		Make any requested changes :

	for ( int i = 0; i < RULER_NUM_HOTKEYS; i++ ) {
		if ( NewHotkey[i] != -1 )
			Hotkey[i] = NewHotkey[i];

		if ( NewHotkeyMode[i] != -1 )
			HotkeyMode[i] = NewHotkeyMode[i];
	}

	if ( NewBackColor != -1 )
		BackColor = NewBackColor;

	if ( NewHashColor != -1 )
		HashColor = NewHashColor;

	if ( NewAlpha != -1 )
		Alpha = NewAlpha;

	if ( NewWidth != -1 )
		PenWidth = NewWidth;

	if ( NewWidth2 != -1 )
		PenWidth2 = NewWidth2;

	if ( NewPosition != -1 )
		Position = NewPosition;

	if ( NewTopOffset != -1 )
		TopOffset = NewTopOffset;

	if ( NewStartDisabled != -1 )
		DisabledAtStartup = NewStartDisabled;

	if ( NewTranspMode != -1 )
		TranspMode = NewTranspMode;

	for ( int i = 0; i < 16; i++ )
		if ( NewRulerColor[i] != -1 )
			RulerColorForBG[i] = NewRulerColor[i];

	if ( NewShow == 1 )
		Disabled = FALSE;
	else if ( NewShow == 0 )
		Disabled = TRUE;

	if ( NewHookListCmd != -1 )
		HookListCmd = NewHookListCmd;


	if ( RedrawNeeded )
		RedrawRulers();


	if ( SaveKey || ( NewStartDisabled != -1 ) || ( NewHookListCmd != -1 ) )
		CondenseHotkeys();


	//		Display the current settings :

	if ( !Quiet ) {

		Printf( L"\r\n%s v%u.%u.%u options:\r\n\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

		for ( int i = 0; i < RULER_NUM_HOTKEYS; i++ ) {
			if ( ( i == 0 ) || Hotkey[i] ) {
				GetKeyName( Hotkey[i], shortbuf, TRUE );
				Printf( L"    Ruler hotkey #%u  . . . . . /K%u: %s\r\n", i + 1, i + 1, shortbuf );
				Printf( L"    Hotkey mode  #%u  . . . . . /M%u: %u \t\t(%s)\r\n", i + 1, i + 1, HotkeyMode[i], HotkeyModeStr[HotkeyMode[i]] );
			}
		}

		Printf( L"    Transparency mode  . . . . /X : %u \t\t(%s)\r\n", TranspMode, TranspModeStr[TranspMode] );

		if ( TranspMode ) {
			W3CColorName( RGB2ColorRef( BackColor ), shortbuf, SHORT_BUF_LEN, FALSE );
			Printf( L"    Background color . . . . . /B : %s \t(%u,%u,%u)\r\n", shortbuf, GetRValue( BackColor ), GetGValue( BackColor ), GetBValue( BackColor ) );
			W3CColorName( RGB2ColorRef( HashColor ), shortbuf, SHORT_BUF_LEN, FALSE );
			Printf( L"    Hash marks color . . . . . /H : %s \t(%u,%u,%u)\r\n", shortbuf, GetRValue( HashColor ), GetGValue( HashColor ), GetBValue( HashColor ) );
			Printf( L"    Ruler alpha  . . . . . . . /A : %u\r\n", Alpha );
		}
		else {
			W3CColorName( RGB2ColorRef( RulerColorForBG[ConsoleBG] ), shortbuf, SHORT_BUF_LEN, FALSE );
			Printf( L"    Ruler color  . . . . . . . /C : %s \t(%u,%u,%u)\r\n", shortbuf, GetRValue( RulerColorForBG[ConsoleBG] ), GetGValue( RulerColorForBG[ConsoleBG] ), GetBValue( RulerColorForBG[ConsoleBG] ) );
		}

		if ( PenWidth == PenWidth2 )
			Printf( L"    Hash marks width . . . . . /W : %u\r\n", PenWidth );
		else
			Printf( L"    Hash marks width . . . . . /W : %u, %u\r\n", PenWidth, PenWidth2 );

		if ( NumAltColumns ) {
			W3CColorName( RGB2ColorRef( AltColumnsColor ), shortbuf, SHORT_BUF_LEN, FALSE );
			Printf( L"    Alt. color columns . . . . /I : %s \t", shortbuf );

			if ( ( AltColumns[0] & RULER_FLAG_MARK_REPEATS ) == 0 )
				QPuts( L"at " );

			for ( unsigned int i = 0; i < NumAltColumns; i++ ) {
				if ( AltColumns[i] & RULER_FLAG_MARK_REPEATS )
					Printf( L"%severy %s column", i ? L", " : L"", OrdinalString( AltColumns[i] & ~RULER_FLAG_MARK_REPEATS ) );
				else
					Printf( L"%s%u", i ? L", " : L"", AltColumns[i] );
			}
		
			CrLf();
		}

		if ( NumAltRows ) {
			W3CColorName( RGB2ColorRef( AltRowsColor ), shortbuf, SHORT_BUF_LEN, FALSE );
			Printf( L"    Alt. color rows  . . . . . /J : %s \t", shortbuf );

			if ( ( AltRows[0] & RULER_FLAG_MARK_REPEATS ) == 0 )
				QPuts( L"at " );

			for ( unsigned int i = 0; i < NumAltRows; i++ ) {
				if ( AltRows[i] & RULER_FLAG_MARK_REPEATS )
					Printf( L"%severy %s row", i ? L", " : L"", OrdinalString( AltRows[i] & ~RULER_FLAG_MARK_REPEATS ) );
				else
					Printf( L"%s%u", i ? L", " : L"", AltRows[i] );
			}
		
			CrLf();
		}

		Printf( L"    Ruler position . . . . . . /P : %u \t\t(%s)\r\n", Position, PositionStr[Position] );
		Printf( L"    Currently enabled  . . . . /E : %u \t\t(%s)\r\n", Disabled ? 0 : 1, Disabled ? L"No" : L"Yes" );
		Printf( L"    Hidden at startup  . . . . /Z : %u *\t\t(%s)\r\n", DisabledAtStartup, DisabledAtStartup ? L"Yes" : L"No" );
		Printf( L"    Intercept LIST . . . . . . /U : %u *\t\t(%s)\r\n", HookListCmd, HookListCmd ? L"Yes" : L"No" );

		CrLf();
		Printf( L"    * These settings take effect the next time the plugin is loaded.\r\n" );

	}


	//		Save the current settings if desired :

	if ( SaveKey ) {
		if ( trv = SaveOptions( ( SaveKey > 1 ) ? HKEY_LOCAL_MACHINE : HKEY_CURRENT_USER ) ) {
			GetErrorText( trv, shortbuf );
			CrLf();
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error saving settings to the registry (%s\\%s) :\r\n", RootKeyName( SaveKey ), PluginRegKey );
			Qprintf( ERRHANDLE, L"%s\r\n", shortbuf );
			rv = 2;
		}
		else if ( !Quiet )
			Printf( L"\r\nSettings saved to the registry (%s\\%s).\r\n", RootKeyName( SaveKey ), PluginRegKey );
	}

	if ( !SaveKey && ( ( NewStartDisabled != -1 ) || ( NewHookListCmd != -1 ) ) ) {
		if ( trv = SaveStartupOptions( HKEY_CURRENT_USER, NewStartDisabled, NewHookListCmd ) ) {
			GetErrorText( trv, shortbuf );
			CrLf();
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error saving restart options to the registry (HKCU\\%s) :\r\n", PluginRegKey );
			Qprintf( ERRHANDLE, L"%s\r\n", shortbuf );
			rv = 2;
		}
		else if ( !Quiet )
			Printf( L"\r\nStartup settings saved to the registry (HKCU\\%s).\r\n", PluginRegKey );
	}


	return rv;
}



int RulerHotkey( DWORD Mode )
{
	switch ( Mode ) {

	case 1 :
	case 2 :
	case 3 :

		if ( Disabled )
			Position = RULER_POSITION_TOP, Disabled = FALSE;
		else if ( Position < Mode )
			Position++;
		else
			Disabled = TRUE;

		break;

	case 4 :
	case 5 :
	case 6 :

		Disabled = FALSE;
		Position++;

		if ( Position > ( Mode - 3 ) )
			Position = RULER_POSITION_TOP;

		break;

	case 7 :

		if ( !Disabled && ( Position == RULER_POSITION_LEFT ) ) {
			Disabled = TRUE;
			break;
		}

	case 8 :

		if ( Position )
			Position = RULER_POSITION_TOP;
		else
			Position = RULER_POSITION_LEFT;

		Disabled = FALSE;
		break;

	case 9 :
	case 10 :

		Disabled = Mode - 9;
		break;

	case 11 :
	case 12 :
	case 13 :
	case 14 :

		Position = Mode - 11;
		Disabled = FALSE;
		break;

	case 15 :

		if ( !Disabled && ( Position == RULER_POSITION_RIGHT ) ) {
			Disabled = TRUE;
			break;
		}

	case 16 :

		if ( Position != RULER_POSITION_LEFT )
			Position = RULER_POSITION_LEFT;
		else
			Position = RULER_POSITION_RIGHT;

		Disabled = FALSE;
		break;

	case 17 :
		
		TranspMode = !TranspMode;
		break;

	default :

		Disabled = !Disabled;
	}

	RedrawRulers();
	return 0;
}



DLLExports int WINAPI key( LPKEYINFO24 ki )
{
	int inKey		= ki->nKey;
	int ShiftKeys	= 0;


	if ( NewKeyHook )
		if ( !inKey && ( ki->pszKey ) )
			inKey = ParseKeyName24( ki->pszKey );


	if ( GetAsyncKeyState( VK_SHIFT ) & 0x8000 )
		ShiftKeys |= 1;

	if ( GetAsyncKeyState( VK_CONTROL ) & 0x8000 )
		ShiftKeys |= 2;

	if ( GetAsyncKeyState( VK_MENU ) & 0x8000 )
		ShiftKeys |= 4;

	if ( ShiftKeys == 3 )
		inKey |= 0x20000;
	else if ( ShiftKeys == 6 )
		inKey |= 0x40000;
	else if ( NewKeyHook && ( ShiftKeys == 7 ) )
		inKey |= 0x60000;


	if ( inKey ) {
		for ( int i = 0; i < RULER_NUM_HOTKEYS; i++ ) {
			if ( inKey == Hotkey[i] ) {
				RulerHotkey( HotkeyMode[i] );

				ki->nKey = 0;
				if ( NewKeyHook )
					ki->pszKey[0] = '\0';

				break;
			}
		}
	}


	if ( !inKey && ki->pszLine == NULL ) {

		if ( !IsWindowVisible( ConsoleWindow ) && !TakeCmdRuler ) {
			TakeCmdWindow = FindTakeCmdWindow();

			if ( TakeCmdWindow )
				MyTab = FindMyTab( TakeCmdWindow, ConsoleWindow );
			else
				MyTab = NULL;

			if ( MyTab ) {
				TakeCmdRuler = CreateRulerWindow( RULER_MODE_TCMD, MyTab );

#ifdef DEBUG_CODE

				if ( Debug && TakeCmdRuler )
					_cwprintf( L"   $  Created new TakeCmdRuler = 0x%08X\n", TakeCmdRuler );

#endif	//	DEBUG_CODE

			}
		}

		TopOffset = 0;
		RedrawRulers();
	}

	return 0;
}



DLLExports int WINAPI list( LPTSTR inString )
{
	int NewTopOffset		= 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_FORCE_UPPERCASE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( !wcscmp( arg[i], L"/U" ) )
			NewTopOffset = 3;
	}


	TopOffset = NewTopOffset;
	RedrawRulers();
	return MAGIC_VALUE;
}



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}



DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}



/*

	1.0.0.1		2021-10-02

	1.0.1.0		2021-12-21
				Fixed a problem calling TakeCommandIPC() on 32-bit TCC.

	1.0.2.0		2021-12-22
				No longer links to TakeCommandIPC() dynamically.  That hasn't been
				necessary for a long long time.

	1.0.2.1		2023-06-13
				Removes a call to QueryIsFile(), which seems to have issues in
				TCC 30.0.20.

*/
